import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import powerbi  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()
    auth_group = parser.add_mutually_exclusive_group(required=True)

    parser.add_argument("-t", "--tenant_id", help="PowerBi tenant ID")
    parser.add_argument("-c", "--client_id", help="PowerBi client ID")
    auth_group.add_argument(
        "-s",
        "--secret",
        help="PowerBi password as a string",
    )
    auth_group.add_argument(
        "-cert",
        "--certificate",
        help=(
            "Path to certificate file for authentication. "
            "Accepts: X.509 certificates (.pem, .crt), "
            "PKCS#12 files (.p12, .pfx), or JSON files containing "
            "certificate data or custom authentication secrets."
        ),
    )
    parser.add_argument(
        "-sc",
        "--scopes",
        help="Optional scopes for the Power BI REST API and the Microsoft Graph API",
        nargs="*",
    )
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument("-l", "--login_url", help="Login url (Optional)")
    parser.add_argument(
        "-a", "--api_base", help="Power BI REST API base (Optional)"
    )
    parser.add_argument(
        "-g", "--graph_api_base", help="Microsoft Graph API base (Optional)"
    )

    powerbi.extract_all(**parse_filled_arguments(parser))



from .users import UserDataProcessor

_METADATA = [
    {
        "id": "workspace-one-user-one-group",
        "users": [
            {"graphId": "U1", "displayName": "jean", "principalType": "User"},
            {"graphId": "GR1", "displayName": "GR1", "principalType": "Group"},
        ],
    },
    {
        "id": "workspace-one-user-duplicate-group",
        "users": [
            {"graphId": "U2", "displayName": "michel", "principalType": "User"},
            {"graphId": "GR1", "displayName": "GR1", "principalType": "Group"},
            {"graphId": "?", "displayName": "tata", "principalType": "?"},
        ],
    },
    {"id": "workspace-no-users-field"},
    {"id": "workspace-empty-users-field", "users": []},
]

_GR1_MEMBERS = [
    {"id": "U3", "displayName": "jacques", "mail": "jacques@catylog.com"},
    {"id": "U1", "displayName": "toto", "mail": ""},  # duplicate
]


def test_parse_metadata_and_combine_users():
    processor = UserDataProcessor()
    processor.parse_metadata(_METADATA)

    assert {u["graphId"] for u in processor.users} == {"U1", "U2"}

    combined_users = processor.combine_users(_GR1_MEMBERS)
    assert {u["graphId"] for u in combined_users} == {"U1", "U2", "U3"}


def test__normalize_group_member():
    processor = UserDataProcessor()
    member = _GR1_MEMBERS[0]
    expected_user = {
        "graphId": "U3",
        "displayName": "jacques",
        "emailAddress": "jacques@catylog.com",
    }
    assert processor._normalize_group_member(member) == expected_user



from pydantic.dataclasses import dataclass

from ....warehouse.bigquery import BigQueryCredentials


@dataclass
class CountCredentials(BigQueryCredentials):
    """Count credentials extending BigQuery credentials with additional dataset information"""

    dataset_id: str



from .assets import PowerBiAsset
from .client import (
    CLIENT_APP_BASE,
    DEFAULT_SCOPES,
    GRAPH_API_BASE_PATH,
    REST_API_BASE_PATH,
    MicrosoftGraphPIClient,
    PowerbiClient,
    PowerbiCredentials,
)
from .extract import extract_all
from .users import consolidate_powerbi_users



from collections.abc import Iterator

from ....utils import empty_iterator
from ....warehouse.abstract import WarehouseAsset
from ....warehouse.bigquery import BigQueryClient, BigQueryQueryBuilder
from .. import LookerStudioAsset
from .admin_sdk_client import USER_EMAIL_FIELD, AdminSDKClient
from .credentials import LookerStudioCredentials
from .looker_studio_api_client import LookerStudioAPIClient

_ASSERT_ADMIN_SDK_CLIENT_MSG = "Missing Admin SDK client."
_ASSERT_BIGQUERY_CLIENT_MSG = "Missing BigQuery client."
_ASSERT_LOOKER_STUDIO_CLIENT_MSG = "Missing Looker Studio client."


class LookerStudioQueryBuilder(BigQueryQueryBuilder):
    def job_history_queries(self) -> list:
        """
        This class and method are a convenient workaround to build the
        ExtractionQueries which retrieve BigQuery's job history, but filtered on
        Looker Studio only.

        Compared to the generic BigQuery query history, only the SQL "template"
        changes. By defining this class here, this will pick the SQL file
        `queries/query.sql` located in the same directory as this file.
        """
        return super().build(WarehouseAsset.QUERY)  # type: ignore


class LookerStudioClient:
    """
    Acts as a wrapper class to fetch Looker Studio assets, which requires
    coordinating calls between
    - the Admin SDK API
    - the Looker Studio API
    - BigQuery (for source queries)

    All credentials are optional because the package command allows
    enabling/disabling specific extractions. For example, the user may not allow
    accessing Admin endpoints.

    See the following Notion page for more details about Domain-Wide Delegation
    and the different command line args:
    https://www.notion.so/castordoc/Looker-Studio-extract-source-queries-only-no-DWD-required-2aba1c3d458580d78199da6ca4c7f46f?source=copy_link
    """

    def __init__(
        self,
        credentials: LookerStudioCredentials | None = None,
        bigquery_credentials: dict | None = None,
        user_emails: list[str] | None = None,
        database_allowed: set[str] | None = None,
        database_blocked: set[str] | None = None,
    ):
        self.admin_sdk_client: AdminSDKClient | None = None
        self.bigquery_client: BigQueryClient | None = None
        self.looker_studio_client: LookerStudioAPIClient | None = None

        self.user_emails = user_emails

        self._initialize_clients(
            credentials=credentials,
            bigquery_credentials=bigquery_credentials,
            db_allowed=database_allowed,
            db_blocked=database_blocked,
        )

    def _initialize_clients(
        self,
        credentials: LookerStudioCredentials | None,
        bigquery_credentials: dict | None,
        db_allowed: set[str] | None,
        db_blocked: set[str] | None,
    ) -> None:
        if credentials:
            self.admin_sdk_client = AdminSDKClient(credentials)
            self.looker_studio_client = LookerStudioAPIClient(credentials)

        if bigquery_credentials:
            self.bigquery_client = BigQueryClient(
                credentials=bigquery_credentials,
                db_allowed=db_allowed,
                db_blocked=db_blocked,
            )

    def _list_user_emails(self) -> Iterator[str]:
        """
        Lists user emails either from a provided JSON file or via the Admin SDK API.

        Using all Google Workspace users can be inefficient for large clients -
        the client might spend hours checking thousands of users for Looker Studio
        assets when only a handful actually own any. A JSON file allows
        targeting known owners instead.
        """
        if self.user_emails is not None:
            yield from self.user_emails
            return

        assert self.admin_sdk_client, _ASSERT_ADMIN_SDK_CLIENT_MSG
        for user in self.admin_sdk_client.list_users():
            yield user[USER_EMAIL_FIELD]

    def _get_assets(self) -> Iterator[dict]:
        """
        Extracts reports and data sources user by user. The loop is necessary
        because the Looker Studio API can only retrieve the assets owned by a
        single user.
        """
        assert self.looker_studio_client, _ASSERT_LOOKER_STUDIO_CLIENT_MSG

        for user_email in self._list_user_emails():
            yield from self.looker_studio_client.fetch_user_assets(user_email)

    def _get_source_queries(self) -> Iterator[dict]:
        """
        Extracts the BigQuery jobs triggered by Looker Studio. The last job
        per data source is returned.
        """
        if not self.bigquery_client:
            return empty_iterator()

        query_builder = LookerStudioQueryBuilder(
            regions=self.bigquery_client.get_regions(),
            datasets=self.bigquery_client.get_datasets(),
            extended_regions=self.bigquery_client.get_extended_regions(),
        )

        queries = query_builder.job_history_queries()

        for query in queries:
            yield from self.bigquery_client.execute(query)

    def fetch(self, asset: LookerStudioAsset) -> Iterator[dict]:
        if asset == LookerStudioAsset.ASSETS:
            yield from self._get_assets()

        elif asset == LookerStudioAsset.SOURCE_QUERIES:
            yield from self._get_source_queries()

        elif asset == LookerStudioAsset.VIEW_ACTIVITY:
            assert self.admin_sdk_client, _ASSERT_ADMIN_SDK_CLIENT_MSG
            yield from self.admin_sdk_client.list_view_events()

        else:
            raise ValueError(f"The asset {asset}, is not supported")



import logging
from collections.abc import Iterator
from datetime import date
from functools import partial
from http import HTTPStatus
from time import sleep

import requests
from requests import HTTPError

from ....utils import (
    APIClient,
    deduplicate,
    fetch_all_pages,
    retry_request,
)
from .authentication import PowerBiBearerAuth
from .constants import Keys
from .credentials import PowerbiCredentials
from .endpoints import PowerBiEndpointFactory
from .pagination import PowerBIAPIPagination

POWERBI_DEFAULT_TIMEOUT_S = 30
# The route we use to fetch workspaces info can retrieve a maximum of
# 100 workspaces per call
# More: https://learn.microsoft.com/en-us/rest/api/power-bi/admin/workspace-info-post-workspace-info#request-body
METADATA_BATCH_SIZE = 100
POWERBI_SCAN_STATUS_DONE = "Succeeded"
POWERBI_SCAN_SLEEP_S = 1
POWERBI_SCAN_TIMEOUT_S = 60

MAX_RETRY_PAGES = 1
RETRY_PAGES_TIMEOUT_MS = 35 * 1000  # 35 seconds

KEYS_TO_HIDE = ("ClientIP", "UserAgent")

logger = logging.getLogger(__name__)


class PowerBIAPIClient(APIClient):
    def __init__(
        self,
        credentials: PowerbiCredentials,
    ):
        auth = PowerBiBearerAuth(credentials=credentials, scope_type="powerbi")
        super().__init__(
            auth=auth,
            timeout=POWERBI_DEFAULT_TIMEOUT_S,
        )
        self.endpoint_factory = PowerBiEndpointFactory(
            login_url=credentials.login_url,
            api_base=credentials.api_base,
        )

    def activity_events(self, day: date | None = None) -> Iterator[dict]:
        """
        Returns a list of activity events for the organization.
        https://learn.microsoft.com/en-us/power-bi/admin/service-admin-auditing#activityevents-rest-api
        - when no day is specified, fallback is yesterday
        """
        request = partial(
            self._get,
            endpoint=self.endpoint_factory.activity_events(day),
        )
        for event in fetch_all_pages(request, PowerBIAPIPagination):
            for key in KEYS_TO_HIDE:
                if key in event:
                    del event[key]
            yield event

    def _dataflow_datasources(self, dataflow_id: str) -> dict | list:
        """
        Extracts the datasources of a dataflow.
        This endpoint sometimes returns 400 and 409 errors.
        Expection logic was added to deal with these cases.
        """
        try:
            return self._get(
                self.endpoint_factory.dataflow_datasources(dataflow_id)
            )[Keys.VALUE]
        except requests.HTTPError as e:
            logger.debug(e)
            logger.info(
                f"Skipping datasources extraction for Dataflow {dataflow_id}"
            )
            logger.info(e.response.text)
            return []

    def _dataflow_definition(self, dataflow_id: str) -> dict | list:
        """
        Extracts the definition of a dataflow.
        This endpoint sometimes returns 400 and 409 errors.
        Expection logic was added to deal with these cases.
        """
        try:
            return self._get(
                self.endpoint_factory.export_dataflow_definition(dataflow_id)
            )
        except requests.HTTPError as e:
            logger.debug(e)
            logger.info(
                f"Skipping definition extraction for Dataflow {dataflow_id}"
            )
            logger.info(e.response.text)
            return []

    def dataflows(self) -> list[dict]:
        """
        Returns a list of dataflows for the organization and their related information.

        Get the list of Dataflows:
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/dataflows-get-dataflows-as-admin

        Get the datasources information of a Dataflow:
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/dataflows-get-dataflow-datasources-as-admin

        Get the definition of a Dataflow:
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/dataflows-export-dataflow-as-admin
        """
        dataflows = self._get(self.endpoint_factory.dataflows())[Keys.VALUE]

        for dataflow in dataflows:
            dataflow_id = dataflow["objectId"]

            dataflow["datasources"] = self._dataflow_datasources(dataflow_id)
            dataflow["definition"] = self._dataflow_definition(dataflow_id)
        return dataflows

    def datasets(self) -> Iterator[dict]:
        """
        Returns a list of datasets for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/datasets-get-datasets-as-admin
        """
        yield from self._get(self.endpoint_factory.datasets())[Keys.VALUE]

    def dashboards(self) -> Iterator[dict]:
        """
        Returns a list of dashboards for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/dashboards-get-dashboards-as-admin
        """
        yield from self._get(self.endpoint_factory.dashboards())[Keys.VALUE]

    @retry_request(
        status_codes=(HTTPStatus.TOO_MANY_REQUESTS,),
        max_retries=MAX_RETRY_PAGES,
        base_ms=RETRY_PAGES_TIMEOUT_MS,
    )
    def _pages(self, report_id: str) -> Iterator[dict]:
        """
        Extracts the pages of a report.
        This endpoint is very flaky and frequently returns 400 and 404 errors.
        After around 50 requests, it hits the rate limit and returns 429 Too Many Requests,
        which is why we retry it after a short delay.
        Timeouts are also common; we must skip them because the extraction task
        might take too long otherwise.
        """
        pages_endpoint = self.endpoint_factory.pages(report_id)
        return self._get(pages_endpoint, retry_on_timeout=False)[Keys.VALUE]

    def _get_accessible_workspace_ids(self) -> set[str]:
        """
        Returns the IDs of workspaces the credentials have access to.
        https://learn.microsoft.com/en-us/rest/api/power-bi/groups/get-groups
        """
        workspaces = self._get(self.endpoint_factory.groups())[Keys.VALUE]
        return {w[Keys.ID] for w in workspaces}

    def _get_pages_or_none(
        self, report: dict, accessible_workspace_ids: set[str]
    ) -> list[dict] | None:
        report_id = report.get(Keys.ID)
        workspace_id = report.get(Keys.WORKSPACE_ID)

        if workspace_id not in accessible_workspace_ids:
            return None

        try:
            return self._pages(report_id)
        except (requests.HTTPError, requests.exceptions.Timeout) as e:
            logger.debug(e)
            return None

    def reports(self) -> list[dict]:
        """
        Returns a list of reports for the organization.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/reports-get-reports-as-admin

        For each report, it also attempts to extracts its pages. Unfortunately,
        the endpoint is not an Admin endpoint and therefore only works on
        reports beloning to workspaces that the credentials have access to.
        This may also take long enough for the token to expire, so it's
        force-refreshed at the end to avoid issues with the next requests.
        https://learn.microsoft.com/en-us/rest/api/power-bi/reports/get-pages
        """
        reports = self._get(self.endpoint_factory.reports())[Keys.VALUE]

        accessible_workspace_ids = self._get_accessible_workspace_ids()
        if not accessible_workspace_ids:
            logger.warning(
                "No accessible workspaces found. Skipping pages extraction."
            )
            return reports

        for report in reports:
            pages = self._get_pages_or_none(report, accessible_workspace_ids)
            if pages:
                listed_pages: list[dict] = pages
                deduplicated_pages = deduplicate("name", listed_pages)
                report["pages"] = deduplicated_pages

        self._auth.refresh_token()
        return reports

    def _workspace_ids(self) -> list[str]:
        """
        Get workspaces ids from powerBI admin API.
        more: https://learn.microsoft.com/en-us/rest/api/power-bi/admin/workspace-info-get-modified-workspaces
        """
        params: dict[str, bool | str] = {
            Keys.INACTIVE_WORKSPACES: True,
            Keys.PERSONAL_WORKSPACES: True,
        }

        response = self._get(
            self.endpoint_factory.workspace_ids(),
            params=params,
        )

        return [x[Keys.ID] for x in response]

    def _get_scan_result(self, scan_id: int) -> Iterator[dict]:
        endpoint = self.endpoint_factory.metadata_scan_result(scan_id)
        yield from self._get(endpoint)[Keys.WORKSPACES]

    def _wait_for_scan_result(self, scan_id: int) -> bool:
        """
        Periodically checks the status of the metadata scan until the results
        are ready.
        """
        endpoint = self.endpoint_factory.metadata_scan_status(scan_id)
        total_waiting_time_s = 0

        while total_waiting_time_s < POWERBI_SCAN_TIMEOUT_S:
            try:
                result = self._get(endpoint)
            except HTTPError as e:
                logger.error(f"Scan {scan_id} failed. Error: {e}")
                return False

            if result[Keys.STATUS] == POWERBI_SCAN_STATUS_DONE:
                logger.info(f"scan {scan_id} ready")
                return True

            total_waiting_time_s += POWERBI_SCAN_SLEEP_S
            logger.info(
                f"Waiting {POWERBI_SCAN_SLEEP_S} sec for scan {scan_id} to be ready…",
            )
            sleep(POWERBI_SCAN_SLEEP_S)

        logger.warning(f"Scan {scan_id} timed out")
        return False

    def _create_scan(self, workspaces_ids: list[str]) -> int:
        """
        Tells the Power BI API to start an asynchronous metadata scan.
        Returns the scan's ID.
        https://learn.microsoft.com/en-us/rest/api/power-bi/admin/workspace-info-post-workspace-info
        """
        params = {
            "datasetExpressions": True,
            "datasetSchema": True,
            "datasourceDetails": True,
            "getArtifactUsers": True,
            "lineage": True,
        }
        request_body = {"workspaces": workspaces_ids}
        scan_id = self._post(
            self.endpoint_factory.metadata_create_scan(),
            params=params,
            data=request_body,
        )
        return scan_id[Keys.ID]

    def metadata(self) -> Iterator[dict]:
        """
        Fetch metadata by workspace. The metadata scanning is asynchronous and
        requires the following steps:
        - create the asynchronous scan
        - periodically check the scan status to know when it's finished
        - get the actual scan results
        https://learn.microsoft.com/en-us/power-bi/enterprise/service-admin-metadata-scanning
        """
        ids = self._workspace_ids()

        for index in range(0, len(ids), METADATA_BATCH_SIZE):
            batch_ids = ids[index : index + METADATA_BATCH_SIZE]
            scan_id = self._create_scan(batch_ids)
            self._wait_for_scan_result(scan_id)
            yield from self._get_scan_result(scan_id)



from ...utils import ExtractionQuery
from ..abstract import (
    AbstractQueryBuilder,
    TimeFilter,
    WarehouseAsset,
)
from .types import LongQuery

COMMON_CTE_PATH = "sql/long/common_cte.sql"
LONG_QUERY_PATHS: dict[LongQuery, str] = {
    LongQuery.METADATA: "sql/long/query_metadata.sql",
    LongQuery.PARTS: "sql/long/query_parts.sql",
}


class RedshiftQueryBuilder(AbstractQueryBuilder):
    """
    Builds queries to extract assets from Redshift.
    """

    def __init__(
        self,
        is_serverless: bool = False,
        time_filter: TimeFilter | None = None,
    ):
        super().__init__(time_filter=time_filter)
        self.is_serverless = is_serverless

    def build_custom_query(self, query_file: str) -> ExtractionQuery:
        """
        Build custom queries to extract redshift metadata that does not directly
        translate to typical assets (typical asset = database, schema, user...)
        """
        statement = self._load_from_file(query_file)
        params = self._time_filter.to_dict()
        return ExtractionQuery(statement, params)

    def build_long_query(self, long_query_type: LongQuery) -> ExtractionQuery:
        common_cte = self._load_from_file(COMMON_CTE_PATH)
        specific = self._load_from_file(LONG_QUERY_PATHS[long_query_type])
        statement = "\n".join([common_cte, specific])
        params = self._time_filter.to_dict()
        return ExtractionQuery(statement, params)

    def build(self, asset: WarehouseAsset) -> list[ExtractionQuery]:
        if asset == WarehouseAsset.QUERY and self.is_serverless:
            # To get the query history in Redshift Serverless, we cannot use STL tables
            query = self.build_custom_query("sql_serverless/query.sql")
        elif asset == WarehouseAsset.QUERY:
            # the default build expects `query.sql` one level higher
            query = self.build_custom_query("sql/short/query.sql")
        else:
            query = self.build_default(asset)
        return [query]



import logging

import tableauserverclient as TSC  # type: ignore

from ....utils import SerializedAsset
from ..assets import TABLEAU_PULSE_ASSETS, TableauAsset
from ..constants import CREDENTIALS_SITE_ID_KEY, DEFAULT_TIMEOUT_SECONDS
from .client_metadata_api import TableauClientMetadataApi
from .client_rest_api import TableauClientRestApi
from .client_tsc import TableauClientTSC
from .credentials import TableauCredentials
from .gql_queries import LIGHT_GQL_QUERIES

_TABLEAU_EXTRACT_LABELS = (
    "[extract].[extract]",
    "[public].[extract]",
)

logger = logging.getLogger(__name__)

# these assets must be extracted via TableauServerClient (TSC)
_TSC_ASSETS = (
    # projects are not available in Metadata API
    TableauAsset.PROJECT,
    # view count are not available in Metadata API
    TableauAsset.USAGE,
    # only users who published content can be extracted from MetadataAPI
    TableauAsset.USER,
)

# these assets must be extracted via the REST API
_REST_API_ASSETS = (
    # Tableau Pulse assets are only available in REST API
    TableauAsset.METRIC,
    TableauAsset.METRIC_DEFINITION,
    TableauAsset.SUBSCRIPTION,
)

logging.getLogger("tableau.endpoint").setLevel(logging.WARNING)


def _add_site_id(workbooks: SerializedAsset, site_id: str) -> SerializedAsset:
    """
    Add site_id from credentials: it's necessary to compute workbook's url
    """
    for workbook in workbooks:
        workbook[CREDENTIALS_SITE_ID_KEY] = site_id
    return workbooks


def _merge_datasources(
    datasources: SerializedAsset,
    tsc_datasources: SerializedAsset,
) -> SerializedAsset:
    """
    Enrich datasources with fields coming from TableauServerClient:
    - project_luid
    - webpage_url
    """

    mapping = {row["id"]: row for row in tsc_datasources}

    for datasource in datasources:
        if datasource["__typename"] != "PublishedDatasource":
            # embedded datasources are bound to workbooks => no project
            # embedded datasources cannot be accessed via URL => no webpage_url
            continue
        luid = datasource["luid"]
        tsc_datasource = mapping.get(luid)
        if not tsc_datasource:
            # it happens that a datasource is in Metadata API but not in TSC
            datasource["projectLuid"] = None
            datasource["webpageUrl"] = None
            continue
        datasource["projectLuid"] = tsc_datasource["project_id"]
        datasource["webpageUrl"] = tsc_datasource["webpage_url"]

    return datasources


def _merge_workbooks(
    workbooks: SerializedAsset,
    tsc_workbooks: SerializedAsset,
) -> SerializedAsset:
    """
    Enrich workbooks with fields coming from TableauServerClient:
    - project_luid
    """

    mapping = {row["id"]: row for row in tsc_workbooks}

    for workbook in workbooks:
        luid = workbook["luid"]
        tsc_workbook = mapping.get(luid)
        if not tsc_workbook:
            # it happens that a workbook is in Metadata API but not in TSC
            # in this case, we push the workbook with default project
            logger.warning(f"Workbook {luid} was not found in TSC")
            workbook["projectLuid"] = None
            continue

        workbook["projectLuid"] = tsc_workbook["project_id"]

    return workbooks


def _server(
    server_url: str,
    timeout_sec: int,
    ignore_ssl: bool = False,
) -> TSC.Server:
    verify = not ignore_ssl
    options = {"verify": verify, "timeout": timeout_sec}
    server = TSC.Server(server_url, use_server_version=True)
    server.add_http_options(options)
    return server


def _keep_table(table: dict) -> bool:
    """
    Determine whether a table should be kept during the initial lightweight
    extraction phase by excluding known Tableau Extract tables.

    Extracts such as “[extract].[extract]” and “[public].[extract]” contribute
    hundreds of thousands of unnecessary columns in some environments and
    cannot be filtered out at the GraphQL level, so they are removed here.
    """
    full_name = table.get("fullName") or ""
    clean_full_name = full_name.strip().lower()
    return clean_full_name not in _TABLEAU_EXTRACT_LABELS


class TableauClient:
    """
    Connect to Tableau's API and extract assets.

    Relies on TableauServerClient (TSC) overlay for authentication
    https://tableau.github.io/server-client-python/docs/
    """

    def __init__(
        self,
        credentials: TableauCredentials,
        timeout_sec: int = DEFAULT_TIMEOUT_SECONDS,
        with_columns: bool = True,
        with_fields: bool = True,
        with_pulse: bool = False,
        override_page_size: int | None = None,
        ignore_ssl: bool = False,
    ):
        self._credentials = credentials
        self._server = _server(
            server_url=credentials.server_url,
            timeout_sec=timeout_sec,
            ignore_ssl=ignore_ssl,
        )
        self._with_columns = with_columns
        self._with_fields = with_fields
        self._with_pulse = with_pulse

        self._client_metadata = TableauClientMetadataApi(
            server=self._server,
            override_page_size=override_page_size,
        )
        self._client_rest = TableauClientRestApi(server=self._server)
        self._client_tsc = TableauClientTSC(server=self._server)

    @property
    def base_url(self) -> str:
        return self._credentials.server_url

    @property
    def name(self) -> str:
        return "Tableau/API"

    def _user_password_login(self) -> None:
        """Login into Tableau using user and password"""
        user = self._credentials.user
        password = self._credentials.password
        site_id = self._credentials.site_id
        if user is None or password is None:
            raise ValueError(
                "Invalid credentials: user/password must be provided"
            )
        auth = TSC.TableauAuth(user, password, site_id=site_id)
        self._server.auth.sign_in(auth)

    def _pat_login(self) -> None:
        """Login into Tableau using personal authentication token"""
        token_name = self._credentials.token_name
        token = self._credentials.token
        site_id = self._credentials.site_id
        if token_name is None or token is None:
            raise ValueError(
                "Invalid credentials: token_name/token must be provided"
            )
        auth = TSC.PersonalAccessTokenAuth(token_name, token, site_id=site_id)
        self._server.auth.sign_in(auth)

    def login(self) -> None:
        """
        Depending on the given credentials, logs-in using either:
        - user/password
        - token_name/value (Personal Access Token)
        https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_concepts_auth.htm

        Raises an error if none can be found
        """

        if self._credentials.user and self._credentials.password:
            logger.info("Logging in using user and password authentication")
            return self._user_password_login()

        if self._credentials.token_name and self._credentials.token:
            logger.info("Logging in using token authentication")
            return self._pat_login()

        raise ValueError(
            "Invalid credentials: either user/password or PAT must be provided",
        )

    def _fetch_datasources(self) -> SerializedAsset:
        asset = TableauAsset.DATASOURCE

        datasources = self._client_metadata.fetch(asset)
        tsc_datasources = self._client_tsc.fetch(asset)

        return _merge_datasources(datasources, tsc_datasources)

    def _fetch_workbooks(self) -> SerializedAsset:
        asset = TableauAsset.WORKBOOK

        site_id = self._credentials.site_id
        workbooks = self._client_metadata.fetch(asset)
        workbooks = _add_site_id(workbooks, site_id)

        workbook_projects = self._client_tsc.fetch(asset)

        return _merge_workbooks(workbooks, workbook_projects)

    def fetch_table_ids(self) -> list[str]:
        """
         Return the list of Tableau table IDs after performing a lightweight,
         pre-filtering extraction.

        This method issues a minimal GraphQL metadata query (id + fullName)
        to retrieve all tables, then filters out Tableau Extract tables such as
        “[extract].[extract]” and “[public].[extract]”. These Extract tables often
        generate extremely large numbers of columns and cannot be excluded at the
        GraphQL query level, so they are removed here in Python.

        The resulting ID list is used as the input for subsequent, deeper
        extraction steps (e.g., fetching columns).

        See exploration:
        https://www.notion.so/castordoc/N_Able-x-Tableau-fix-N_Able-column-lineage-2b7a1c3d4585807e981cf7eec499440c
        """
        logger.info("Fetching table-ids...")
        resource, fields, page_size = LIGHT_GQL_QUERIES[TableauAsset.TABLE]
        all_tables = self._client_metadata.call(
            resource=resource,
            fields=fields,
            page_size=page_size,
            show_progress=False,  # quick pre-extraction, hide logs
        )
        logger.info(f"Retrieved {len(all_tables)} table-ids")
        filtered_table_ids = [
            table["id"] for table in all_tables if _keep_table(table)
        ]
        logger.info(
            f"Keeping {len(filtered_table_ids)} table-ids (Removed 'EXTRACTS')"
        )

        return filtered_table_ids

    def fetch_column_ids(
        self,
        table_ids: list[str],
    ) -> SerializedAsset:
        """
        Fetch and return all column IDs for the given list of table IDs.
        """
        logger.info("Fetching column-ids...")
        resource, fields, page_size = LIGHT_GQL_QUERIES[TableauAsset.COLUMN]
        tables = self._client_metadata.call(
            resource=resource,
            fields=fields,
            page_size=page_size,
            filter_ids=table_ids,
            show_progress=False,  # quick pre-extraction, hide logs
        )
        column_ids = [
            column["id"] for table in tables for column in table["columns"]
        ]
        logger.info(
            f"Retrieved {len(column_ids)} column-ids (Removed 'EXTRACTS')"
        )
        return column_ids

    def fetch(
        self,
        asset: TableauAsset,
        filter_ids: list[str] | None = None,
    ) -> SerializedAsset:
        """
        Extract the given Tableau Asset
        """
        if asset in TABLEAU_PULSE_ASSETS and not self._with_pulse:
            logger.info(f"Skipping asset {asset} - Tableau Pulse de-activated")
            return []

        if asset == TableauAsset.COLUMN and not self._with_columns:
            logger.info(f"Skipping asset {asset} - deactivated columns")
            return []

        if asset == TableauAsset.FIELD and not self._with_fields:
            logger.info(f"Skipping asset {asset} - deactivated fields")
            return []

        logger.info(f"Extracting {asset.name}...")

        if asset == TableauAsset.DATASOURCE:
            # two APIs are required to extract datasources
            return self._fetch_datasources()

        if asset == TableauAsset.WORKBOOK:
            # two APIs are required to extract workbooks
            return self._fetch_workbooks()

        if asset in _TSC_ASSETS:
            # some assets can only be extracted via TSC
            return self._client_tsc.fetch(asset)

        if asset in _REST_API_ASSETS:
            # some assets can only be extracted via REST API
            return self._client_rest.fetch(asset)

        # other assets can be extracted via Metadata API
        return self._client_metadata.fetch(
            asset=asset,
            filter_ids=filter_ids,
        )



import json
import logging
from collections.abc import Iterable

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import PowerBiAsset
from .client import PowerbiCertificate, PowerbiClient, PowerbiCredentials
from .client.graph_api_client import MicrosoftGraphAccessForbidden
from .users import consolidate_powerbi_users

logger = logging.getLogger(__name__)


_ASSETS_SUBSET = (
    PowerBiAsset.DATAFLOWS,
    PowerBiAsset.DATASETS,
    PowerBiAsset.DASHBOARDS,
    PowerBiAsset.REPORTS,
    PowerBiAsset.ACTIVITY_EVENTS,
)


def _load_certificate(
    certificate: str | None,
) -> PowerbiCertificate | None:
    if not certificate:
        return None

    with open(certificate) as file:
        cert = json.load(file)
        return PowerbiCertificate(**cert)


def iterate_all_data(
    client: PowerbiClient,
) -> Iterable[tuple[PowerBiAsset, list | dict]]:
    """
    Extracts Power BI data. Metadata and users require special handling.
    The user extraction is skipped if the user does not have access
    to Microsoft Graph.
    """
    for asset in _ASSETS_SUBSET:
        logger.info(f"Extracting {asset.name} from API")
        data = list(deep_serialize(client.fetch(asset)))
        yield asset, data
        logger.info(f"Extracted {len(data)} {asset.name} from API")

    logger.info(f"Extracting {PowerBiAsset.METADATA} from API")
    metadata = list(client.fetch(PowerBiAsset.METADATA))
    yield PowerBiAsset.METADATA, deep_serialize(metadata)

    try:
        combined_users = consolidate_powerbi_users(client, metadata)
        yield PowerBiAsset.USERS, deep_serialize(list(combined_users))
    except MicrosoftGraphAccessForbidden:
        logger.warning(
            "Could not list group members: "
            "the user does not have permissions to access Microsoft Graph."
        )


def extract_all(**kwargs) -> None:
    """
    Extract data from PowerBI REST API
    Store the output files locally under the given output_directory
    """
    _output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    creds = PowerbiCredentials(
        api_base=kwargs.get("api_base"),
        certificate=_load_certificate(kwargs.get("certificate")),
        client_id=kwargs.get("client_id"),
        graph_api_base=kwargs.get("graph_api_base"),
        login_url=kwargs.get("login_url"),
        scopes=kwargs.get("scopes"),
        secret=kwargs.get("secret"),
        tenant_id=kwargs.get("tenant_id"),
    )

    client = PowerbiClient(creds)
    ts = current_timestamp()
    has_users_file = False

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, data)

        if key == PowerBiAsset.USERS:
            has_users_file = True

    write_summary(
        output_directory=_output_directory,
        ts=ts,
        has_powerbi_group_expansion=has_users_file,
    )



from unittest.mock import Mock, patch

import pytest
from requests import HTTPError

from .authentication import msal
from .credentials import PowerbiCredentials
from .graph_api_client import (
    MicrosoftGraphAccessForbidden,
    MicrosoftGraphPIClient,
)


@pytest.fixture
def mock_msal():
    with patch.object(msal, "ConfidentialClientApplication") as mock_app:
        mock_app.return_value.acquire_token_for_client.return_value = {
            "access_token": "fake_token"
        }
        yield mock_app


@pytest.fixture
def graph_client(mock_msal):
    creds = PowerbiCredentials(
        tenant_id="tenant", client_id="client", secret="pwd"
    )
    return MicrosoftGraphPIClient(creds)


def test__users_in_groups(graph_client):
    mock_response = Mock(status_code=403)
    http_error = HTTPError()
    http_error.response = mock_response

    with patch.object(graph_client, "_get", side_effect=http_error):
        with pytest.raises(MicrosoftGraphAccessForbidden):
            list(graph_client.users_in_groups(["id"]))
